<?php

/**
 * Classe per interagire con l'API REST di gCloud Storage
 *
 * @version 1.1.5
 * @package GCloud-SDK-PHP
 */
/**
 * Dipendiamo dalla HTTP_Request2 di PEAR
 */
require_once 'HTTP/Request2.php';

/**
 * Definizione della classe principale
 * 
 * @author Andrea Brancatelli <abrancatelli@schema31.it>
 * @package gCloud_Storage
 */
class gCloud_Storage {

    /**
     * Attuale versione della libreria
     *
     * @var string
     * @access public
     */
    const VERSION = "gCloud_Storage 1.1.5 [Pear]";

    /**
     * Repository su cui operare
     *
     * @var string 
     * @access public
     */
    public $repositoryName = "";

    /**
     * Chiave di autenticazione
     *
     * @var string 
     * @access public
     */
    public $Authentication = "";

    /**
     * gCloud Storage Base Path (hardcoded a "storage.gcloud.schema31.it)
     *
     * @var string 
     * @access public
     */
    public $gCloudStorageHost = "storage.gcloud.schema31.it";

    /**
     * L'ultimo errore ricevuto
     *
     * @var string 
     * @access public
     */
    public $LastError = "";

    /**
     * Variabili di appoggio contenenti la risposta del server
     */
    public $ResponseBody = "";
    public $ResponseStatus = "";

    /**
     * Numero di tentativi per cui riprovare ad eseguire l'operazione richiesta in caso di errori
     *
     * @var int 
     * @access public
     */
    public $maxAutoRetry = 1;

    /**
     * Contatore delle iterazioni (maxAutoRetry) eseguite.
     * Default 1
     *
     * @var int 
     * @access private
     */
    private $lastAutoRetry = 0;

    /**
     * Tempo di attesa tra un tentativo e l'altro in caso di 'autoRetry' (espesso in microsecondi).
     * default 1 secondo
     *
     * @var int 
     * @access public
     */
    public $autoRetrySleep = 1000000;

    /**
     * Contenitore per i dettagli di un file.
     * Viene inizializzato dal metodo 'getFile' per fare in modo che, a fronte di un download, 
     * sia possibile accedere anche ai metadati del file, oltre che al suo contenuto.
     *
     * @var stdClass
     * @access public
     */
    public $fileDetails = NULL;

    /**
     * script timeout (espesso in secondi)
     *
     * @var int 
     * @access public
     */
    const SCRIPT_TIMEOUT = 300;

    /**
     * Mappatura codici di ritorno
     */
    //private $putStatusCodeOK = array(200);
    private $postStatusCodeOK = array(201);
    private $getStatusCodeOK = array(200);
    private $deleteStatusCodeOK = array(200);

    /**
     * Costruttore della classe gCloud_Storage
     */
    function __construct() {

        // se la configurazione prevede un 'default_socket_timeout' minore di 'SCRIPT_TIMEOUT', usa 'SCRIPT_TIMEOUT'
        if (ini_get('default_socket_timeout') < self::SCRIPT_TIMEOUT) {
            ini_set("default_socket_timeout", self::SCRIPT_TIMEOUT);
        }

        // se la configurazione prevede un 'max_execution_time' minore di 'SCRIPT_TIMEOUT', usa 'SCRIPT_TIMEOUT'
        if (ini_get('max_execution_time') < self::SCRIPT_TIMEOUT) {
            ini_set("max_execution_time", self::SCRIPT_TIMEOUT);
        }
    }

    /**
     * Wrapper generale per tutti i metodi esposti dalla classe.
     * Gestisce le iterazioni 'autoRetry' ed i relativi 'autoRetrySleep' in base al valore ritornato ($response) dal metodo richiesto.
     * Se $response é FALSE, viene eseguito nuovamente il metodo richiesto, fino al raggiungimento del numero massimo di iterazioni 
     * eseguibili, aspettando 'autoRetrySleep' tra un'iterazione e l'altra
     * @return mixed output del metodo chiamato
     * @access public
     */
    public function __call($method, $arguments) {

        $response = FALSE;

        // gestione iterazioni
        while ($response === FALSE && $this->lastAutoRetry <= (int) $this->maxAutoRetry) {

            try {

                // esegue il metodo richiesto
                $response = call_user_func_array(array($this, $method), $arguments);

                if ($response === FALSE) {

                    // esito negativo - prova ad eseguire nuovamente l'operazione richiesta

                    $this->lastAutoRetry++;
                    usleep((int) $this->autoRetrySleep);
                } else {
                    // OK - esce dal ciclo e ritorna l'esito
                    break;
                }
            } catch (Exception $e) {

                // catch eccezione - prova ad eseguire nuovamente l'operazione richiesta

                $response = FALSE;
                $this->LastError = 'Caught exception: ' . $e->getMessage();

                $this->lastAutoRetry++;
                usleep((int) $this->autoRetrySleep);
            }
        }

        return $response;
    }

    /**
     * Carica un documento su gCloud Storage
     *
     * @param string $localFile Path del file da inviare su questo filesystem
     * @param string $mime Mime part del file da inviare (default application/octet-stream)
     * @param string $publicName Nome del file da inviare a gCloud Storage (di default prende il basename di $localFile)
     * @param string $fileKey Riferimento ad una fileKey già esistente per effettuare la "sovrascrittura" di un file
     * @return boolean|string torna la fileKey ritornata da gCloud Storage oppure FALSE in caso di errore
     * @access protected
     */
    protected function sendFile($localFile, $mime = "application/octet-stream", $publicName = FALSE, $fileKey = FALSE) {

        /**
         * Se non ci viene passato un publicName esplicito lo ricaviamo dal nome del file da inviare
         */
        if ($publicName == FALSE) {
            $publicName = basename($localFile);
        }

        /**
         * Verifico se c'è una fileKey preimpostata
         */
        if ($fileKey != FALSE) {
            $extraUrl = "fileKey/" . $fileKey;
        } else {
            $extraUrl = "";
        }

        /**
         * Creiamo un'istanza dell'oggetto Request2 definendo tutti i parametri
         */
        $request = new HTTP_Request2('http://' . $this->gCloudStorageHost . "/api/resource/" . $extraUrl);
        $request->setMethod(HTTP_Request2::METHOD_POST)
                ->setHeader('Authentication', $this->Authentication)
                ->setHeader('repositoryName', $this->repositoryName)
                ->setHeader('User-Agent', self::VERSION)
                ->setHeader('Expect', "")
                ->addUpload('somefile', $localFile, $publicName, $mime);

        /**
         * Catturo la response della richiesta
         */
        $HTTPRequest2Response = $request->send();
        $this->ResponseBody = $HTTPRequest2Response->getBody();
        $this->ResponseStatus = $HTTPRequest2Response->getStatus();

        /**
         * Mi assicuro che mi sia stato tornato un oggetto e che non contenga "error".
         * Questa cosa va fatta più robusta.
         * Comunque intanto ritorno FALSE oppure il fileKey in base alle circostanze
         */
        if (!in_array($this->ResponseStatus, $this->postStatusCodeOK)) {
            // HTTP status code non valido
            $this->LastError = $this->ResponseBody;
            return FALSE;
        } else {

            $response = json_decode($this->ResponseBody);

            if (!is_object($response) || property_exists($response, 'Error')) {
                // Si è verificato un errore
                $this->LastError = $this->ResponseBody;
                return FALSE;
            } else {
                // OK
                return $response->Result[0]->fileKey;
            }
        }
    }

    /**
     * Carica N documenti su gCloud Storage
     * 
     * @param array $localFiles Array contenente i file per cui effettuare l'upload
     * @param string $fileKey Riferimento ad una fileKey già esistente per effettuare la "sovrascrittura" di un file
     * @return boolean|string torna la fileKey ritornata da gCloud Storage oppure FALSE in caso di errore
     * @access protected
     */
    protected function sendMultipleFiles($localFiles = array(), $fileKey = FALSE) {

        // Creiamo un'istanza dell'oggetto Request2 definendo tutti i parametri
        $request = new HTTP_Request2('http://' . $this->gCloudStorageHost . "/api/resource/" . ($fileKey != FALSE ? "fileKey/" . $fileKey : ''));
        $request->setMethod(HTTP_Request2::METHOD_POST)
                ->setHeader('Authentication', $this->Authentication)
                ->setHeader('repositoryName', $this->repositoryName)
                ->setHeader('User-Agent', self::VERSION)
                ->setHeader('Expect', "");

        foreach ($localFiles as $index => $localFile) {

            // estrae il file path
            $localFilePath = (array_key_exists('path', $localFile) ? $localFile['path'] : '');

            // verifica che il file esista
            if (strlen($localFilePath) == 0 || !file_exists($localFilePath)) {
                $this->LastError = 'File not found: ' . $localFilePath;
                return FALSE;
            }

            // Se non ci viene passato un publicName esplicito lo ricaviamo dal nome del file da inviare
            $publicName = ( (array_key_exists('publicName', $localFile) && strlen($localFile['publicName']) > 0) ? $localFile['publicName'] : basename($localFilePath) );

            // Se non ci viene passato un mime type esplicito usiamo il default
            $mime = ( (array_key_exists('mime', $localFile) && strlen($localFile['mime']) > 0) ? $localFile['mime'] : 'application/octet-stream' );

            // Aggiunge all'oggetto HTTP_Request2 il file
            $request->addUpload("fileUpload_$index", $localFilePath, $publicName, $mime);
        }

        /**
         * Catturo la response della richiesta
         */
        $HTTPRequest2Response = $request->send();
        $this->ResponseBody = $HTTPRequest2Response->getBody();
        $this->ResponseStatus = $HTTPRequest2Response->getStatus();

        /**
         * Mi assicuro che mi sia stato tornato un oggetto e che non contenga "error".
         * Questa cosa va fatta più robusta.
         * Comunque intanto ritorno FALSE oppure il fileKey in base alle circostanze
         */
        if (!in_array($this->ResponseStatus, $this->postStatusCodeOK)) {
            // HTTP status code non valido
            $this->LastError = $this->ResponseBody;
            return FALSE;
        } else {

            $response = json_decode($this->ResponseBody);

            if (!is_object($response) || property_exists($response, 'Error')) {
                // Si è verificato un errore
                $this->LastError = $this->ResponseBody;
                return FALSE;
            } else {
                return $response->Result;
            }
        }
    }

    /**
     * Torna le informazioni su un dato file
     * @return mixed torna i dettagli sulla risorsa richiesta oppure FALSE in caso di errore
     * @access protected
     */
    protected function detailFile($fileKey, $fileVersion = FALSE, $chain = NULL, $subFile = NULL) {

        /**
         * Creiamo un'istanza dell'oggetto Request2 definendo tutti i parametri
         */
        $url = 'http://' . $this->gCloudStorageHost . '/api/resource/fileKey/' . $fileKey;

        // se non è stata specificata una versione specifica o è stato passato il carattere '_', usa l'ultima versione disponibile
        $url .= ( ($fileVersion != FALSE && $fileVersion != '_') ? "/fileVersion/$fileVersion" : '/_' );

        // se si sta richedendo un file generato da una chain, è necessario indicare sia il nome della chain che l'indice di output del file
        $url .= ( ($chain != NULL && $subFile != NULL) ? "/chain/$chain/chainFileIndex/$subFile" : '' );

        $request = new HTTP_Request2($url);
        $request->setConfig('follow_redirects', TRUE);
        $request->setConfig('max_redirects', 20);
        $request->setConfig('strict_redirects', TRUE);
        $request->setMethod(HTTP_Request2::METHOD_GET)
                ->setHeader('Authentication', $this->Authentication)
                ->setHeader('repositoryName', $this->repositoryName)
                ->setHeader('User-Agent', self::VERSION);

        /**
         * Catturo la response della richiesta
         */
        $HTTPRequest2Response = $request->send();
        $this->ResponseBody = $HTTPRequest2Response->getBody();
        $this->ResponseStatus = $HTTPRequest2Response->getStatus();

        /**
         * Mi assicuro che mi sia stato tornato un oggetto e che non contenga "error".
         * Questa cosa va fatta più robusta.
         * Comunque intanto ritorno FALSE oppure il fileKey in base alle circostanze
         */
        if (!in_array($this->ResponseStatus, $this->getStatusCodeOK)) {
            // HTTP status code non valido
            $this->LastError = $this->ResponseBody;
            return FALSE;
        } else {

            $response = json_decode($this->ResponseBody);

            if (!is_object($response) || property_exists($response, 'Error')) {
                // Si è verificato un errore
                $this->LastError = $this->ResponseBody;
                return FALSE;
            } else {
                return $response;
            }
        }
    }

    /**
     * Scarica un documento da gCloud Storage
     * @return mixed torna il contenuto della risorsa richiesta oppure FALSE in caso di errore
     * @access protected
     */
    protected function getFile($fileKey, $fileVersion = FALSE, $chain = NULL, $subFile = NULL) {

        /*
         * anche in presenza degli attributi $chain e $subFile, vengono estratte le informazioni relative alla risorsa 'master' in quanto:
         * 1. la risorsa 'master' è utilizzata come riferimento anche per tutte le eventuali risorse generate da una chain
         * 2. la risorsa a cui si sta puntando potrebbe essere generata da una chain di tipo 'OUTPUT', quindi fisicamente non presente a sistema
         */
        $response = $this->detailFile($fileKey, $fileVersion);

        if ($response === FALSE) {
            // file non valido
            return FALSE;
        }

        // rendo accessibile i metadati relativi al file
        $this->fileDetails = $response;

        /**
         * Creiamo un'istanza dell'oggetto Request2 definendo tutti i parametri
         */
        $url = $response->friendlyUrl;

        // il friendlyUrl contiene la fileVersion solo se $fileVersion è stato passato come parametro
        $url .= (!$fileVersion ? "/{$response->fileVersion}" : '');

        // aggiunge all'url i riferimenti alla chain (se necessario)
        $url .= ( ($chain != NULL && $subFile != NULL) ? "/$chain/$subFile" : '' );

        $request = new HTTP_Request2($url);
        $request->setConfig('follow_redirects', TRUE);
        $request->setConfig('max_redirects', 20);
        $request->setConfig('strict_redirects', TRUE);
        $request->setMethod(HTTP_Request2::METHOD_GET)
                ->setHeader('Authentication', $this->Authentication)
                ->setHeader('repositoryName', $this->repositoryName)
                ->setHeader('User-Agent', self::VERSION);

        // CONTROLLO ERRORE:
        $HTTPRequest2Response = $request->send();
        $this->ResponseBody = $HTTPRequest2Response->getBody();
        $this->ResponseStatus = $HTTPRequest2Response->getStatus();

        if (!in_array($this->ResponseStatus, $this->getStatusCodeOK)) {
            // HTTP status code non valido
            $this->LastError = $this->ResponseBody;
            return FALSE;
        } else {
            return $this->ResponseBody;
        }
    }

    /**
     * Cancella un documento da gCloud Storage
     * @return mixed torna i dettagli sulla risorsa cancellata oppure FALSE in caso di errore
     * @access protected
     */
    protected function deleteFile($fileKey, $fileVersion = FALSE) {

        /**
         * Creiamo un'istanza dell'oggetto Request2 definendo tutti i parametri
         */
        $url = 'http://' . $this->gCloudStorageHost . '/api/resource/fileKey/' . $fileKey;
        if ($fileVersion != FALSE) {
            $url .= '/fileVersion/' . $fileVersion;
        }

        $request = new HTTP_Request2($url);

        $request->setConfig('follow_redirects', TRUE);
        $request->setConfig('max_redirects', 20);
        $request->setConfig('strict_redirects', TRUE);
        $request->setMethod(HTTP_Request2::METHOD_DELETE)
                ->setHeader('Authentication', $this->Authentication)
                ->setHeader('repositoryName', $this->repositoryName)
                ->setHeader('User-Agent', self::VERSION);

        /**
         * Catturo la response della richiesta
         */
        $HTTPRequest2Response = $request->send();
        $this->ResponseBody = $HTTPRequest2Response->getBody();
        $this->ResponseStatus = $HTTPRequest2Response->getStatus();

        /**
         * Mi assicuro che mi sia stato tornato un oggetto e che non contenga "error".
         * Questa cosa va fatta pi<9d> robusta.
         * Comunque intanto ritorno FALSE oppure il fileKey in base alle circostanze
         */
        if (!in_array($this->ResponseStatus, $this->deleteStatusCodeOK)) {
            // HTTP status code non valido
            $this->LastError = $this->ResponseBody;
            return FALSE;
        } else {

            $response = json_decode($this->ResponseBody);

            if (!is_object($response) || property_exists($response, 'Error')) {
                // Si è verificato un errore
                $this->LastError = $this->ResponseBody;
                return FALSE;
            } else {
                return $response;
            }
        }
    }

    /**
     * Ritorna lo short-url per accedere ad una risorsa memorizzata all'interno di un repository privato
     * @param string $fileKey
     * @param int $fileVersion (facoltativo - se omesso ritorna l'ultima versione per la fileKey passata come parametro)
     * @param int $expiresAt (facoltativo - indica il numero di secondi per cui il link generato sarà valido)
     * @param string $whitelist (facoltativo - lista di indirizzi IP, separati da punto e virgola, abilitati ad accedere al link generato )
     * @return boolean|string url della risorsa oppure FALSE in caso di errore
     * @access protected
     */
    protected function getShortUrl($fileKey, $fileVersion = NULL, $expiresAt = NULL, $whitelist = NULL) {

        // Creiamo un'istanza dell'oggetto Request2 definendo tutti i parametri
        $request = new HTTP_Request2('http://' . $this->gCloudStorageHost . "/api/shorturl");
        $request->setMethod(HTTP_Request2::METHOD_POST)
                ->setHeader('Authentication', $this->Authentication)
                ->setHeader('repositoryName', $this->repositoryName)
                ->setHeader('fileKey', $fileKey)
                ->setHeader('User-Agent', self::VERSION)
                ->setHeader('Expect', "");

        // è stata indicata una fileVersion?
        if (!is_null($fileVersion)) {
            $request->setHeader('fileVersion', $fileVersion);
        }

        // è stata indicata una durata massima per la validità dell'url?
        if (!is_null($expiresAt)) {
            $request->setHeader('expiresAt', $expiresAt);
        }

        // è stata indicata una white-list di indirizzi IP, i quali potranno accedere alla risorsa?
        if (!is_null($whitelist)) {
            $request->setHeader('whitelist', $whitelist);
        }

        // Catturo la response della richiesta
        $HTTPRequest2Response = $request->send();
        $this->ResponseBody = $HTTPRequest2Response->getBody();
        $this->ResponseStatus = $HTTPRequest2Response->getStatus();

        /**
         * Mi assicuro che mi sia stato tornato un oggetto e che non contenga "error".
         * Questa cosa va fatta più robusta.
         * Comunque intanto ritorno FALSE oppure il fileKey in base alle circostanze
         */
        if (!in_array($this->ResponseStatus, $this->postStatusCodeOK)) {
            // HTTP status code non valido
            $this->LastError = $this->ResponseBody;
            return FALSE;
        } else {
            $response = json_decode($this->ResponseBody);

            if (!is_object($response) || property_exists($response, 'Error')) {
                // Si è verificato un errore
                $this->LastError = $this->ResponseBody;
                return FALSE;
            } else {
                return $response->shortUrl;
            }
        }
    }

}
